using System;
using System.IO;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Text;
using System.Threading;

using Xilium.CefGlue;
using Xilium.CefGlue.Platform;

namespace Test
{
    class CrashRendererProcessResourceHandler : CefResourceHandler
    {
        Stream stream;
        bool delay = false;

        protected override void Cancel()
        {
            if (this.stream != null)
            {
                this.stream.Dispose();
                this.stream = null;
            }
        }

        protected override bool CanGetCookie(CefCookie cookie)
        {
            return true;
        }

        protected override bool CanSetCookie(CefCookie cookie)
        {
            return true;
        }

        protected override bool ProcessRequest(CefRequest request, CefCallback callback)
        {
            Uri uri;

            if (Uri.TryCreate(request.Url, UriKind.Absolute, out uri) == false)
            {
                callback.Continue();
                return true;
            }

            if (uri.PathAndQuery.StartsWith("/delay.gif", StringComparison.InvariantCultureIgnoreCase))
            {
                this.delay = true;

                this.stream = new MemoryStream(Convert.FromBase64String("R0lGODlhAQABAIAAAAUEBAAAACwAAAAAAQABAAACAkQBADs="));
            }
            else
            {
                this.stream = new MemoryStream();

                // Init the stream writer
                StreamWriter sw = new StreamWriter(this.stream);

                // Write the error page
                sw.Write("<!DOCTYPE html>\r\n");
                sw.Write("<html>\r\n");
                sw.Write("    <body>\r\n");
                for (int i = 0; i < 10; i++)
                {
                    sw.Write("<img src=\"/delay.gif?" + i.ToString() + "-" + DateTime.Now.ToFileTimeUtc().ToString() + "\">");
                }
                sw.Write("    </body>\r\n");
                sw.Write("</html>\r\n");
                sw.Flush();
            }

            // Reset the position
            stream.Position = 0;

            return true;
        }

        protected override void GetResponseHeaders(CefResponse response, out long responseLength, out string redirectUrl)
        {
            NameValueCollection responseHeaders = new NameValueCollection();

            response.Status = 200;
            response.StatusText = "OK";
            
            if (this.delay == true)
            {
                responseHeaders.Add("Content-Type", "image/gif");
                response.MimeType = "image/gif";
            }
            else
            {
                responseHeaders.Add("Content-Type", "text/html; charset=utf-8");
                response.MimeType = "text/html";
            }

            response.SetHeaderMap(responseHeaders);

            responseLength = this.stream.Length;
            redirectUrl = null;
        }

        protected override bool ReadResponse(Stream response, int bytesToRead, out int bytesRead, CefCallback callback)
        {
            // Check if datas are available
            if (bytesToRead == 0 || this.stream == null || this.stream.Position >= this.stream.Length)
            {
                bytesRead = 0;
                return false;
            }
            else
            {
                if (this.delay == true)
                {
                    Thread.Sleep(1000);
                }

                byte[] buffer = new byte[bytesToRead];

                // Read from the buffer
                bytesRead = this.stream.Read(buffer, 0, buffer.Length);

                // Write the buffer
                response.Write(buffer, 0, bytesRead);

                // Check if the end of read stream have been reached
                if (this.stream.Length == this.stream.Position)
                {
                    this.stream.Close();
                }

                return true;
            }
        }
    }
}
