﻿using System;
using System.Windows.Forms;
using Cef3Test.Helper;
using Xilium.CefGlue;

namespace Cef3Test.Browser
{
	public sealed class CefWebBrowser : Control
	{
		private bool _handleCreated;

		private WebBrowserCore _core;
		private IntPtr _browserWindowHandle;

		public CefWebBrowser()
		{
			SetStyle(
				ControlStyles.ContainerControl
				| ControlStyles.ResizeRedraw
				| ControlStyles.FixedWidth
				| ControlStyles.FixedHeight
				| ControlStyles.StandardClick
				| ControlStyles.UserMouse
				| ControlStyles.SupportsTransparentBackColor
				| ControlStyles.StandardDoubleClick
				| ControlStyles.OptimizedDoubleBuffer
				| ControlStyles.CacheText
				| ControlStyles.EnableNotifyMessage
				| ControlStyles.DoubleBuffer
				| ControlStyles.OptimizedDoubleBuffer
				| ControlStyles.UseTextForAccessibility
				| ControlStyles.Opaque,
				false);

			SetStyle(
				ControlStyles.UserPaint
				| ControlStyles.AllPaintingInWmPaint
				| ControlStyles.Selectable,
				true);

			_core = new WebBrowserCore(this);
			_core.Created += BrowserCreated;
			_core.Loaded += BrowserLoaded;
		}

		private void BrowserCreated(object sender, EventArgs e)
		{
			_browserWindowHandle = _core.CefBrowser.GetHost().GetWindowHandle();
			ResizeWindow(_browserWindowHandle, Width, Height);
		}

		protected override void OnHandleCreated(EventArgs e)
		{
			base.OnHandleCreated(e);

			var windowInfo = new CefWindowInfo();
			windowInfo.Parent = Handle;

			_core.Create(windowInfo);

			_handleCreated = true;
		}

		protected override void Dispose(bool disposing)
		{
			if (_core != null)
			{
				var browser = _core.CefBrowser;
				var host = browser.GetHost();
				host.CloseBrowser();
				host.ParentWindowWillClose();
				host.Dispose();
				browser.Dispose();
				browser = null;
				_browserWindowHandle = IntPtr.Zero;
			}

			base.Dispose(disposing);
		}
		
		protected override void OnResize(EventArgs e)
		{
			base.OnResize(e);

			var form = TopLevelControl as Form;
			if (form != null && form.WindowState != FormWindowState.Minimized)
			{
				ResizeWindow(_browserWindowHandle, Width, Height);
			}
		}
		
		private static void ResizeWindow(IntPtr handle, int width, int height)
		{
			if (handle != IntPtr.Zero)
			{
				NativeMethods.SetWindowPos(handle, IntPtr.Zero, 0, 0, width, height,
				                           SetWindowPosFlags.NoMove | SetWindowPosFlags.NoZOrder);
			}
		}

		public void Navigate(string url)
		{
			var browser = _core.CefBrowser;

			browser.StopLoad();
			browser.GetMainFrame().LoadUrl(url);
		}
		
		private void BrowserLoaded(object sender, EventArgs e)
		{
			OnLoaded();
		}

		#region event handling methods

		public event EventHandler Loaded;
		
		private void OnLoaded()
		{
			var handler = Loaded;
			if (handler != null)
			{
				handler(this, EventArgs.Empty);
			}
		}

		#endregion

		public WebBrowserCore BrowserCore
		{
			get { return _core; }
		}
	}
}
