﻿using System;
using System.Diagnostics;
using Xilium.CefGlue;

namespace Cef3Test.Browser
{
	public sealed class WebBrowserCore
	{
		private readonly object _owner;
		private readonly CefBrowserSettings _settings;
		private CefClient _client;
		private CefBrowser _browser;
		private bool _created;

		public WebBrowserCore(object owner)
		{
			_owner = owner;
			_settings = new CefBrowserSettings()
				{
					//ImageLoadDisabled = true,
					/*
					 The following settings do NOT work, using command-line switches in CefApp::OnBeforeCommandLineProcessing instead
					 */
					WebSecurityDisabled = true,
					XssAuditorEnabled = false
				};
		}

		public CefBrowser CefBrowser
		{
			get { return _browser; }
		}

		public void Create(CefWindowInfo windowInfo)
		{
			if (_client == null)
			{
				_client = new WebClient(this);
			}

			CefBrowserHost.CreateBrowser(windowInfo, _client, _settings, "about:blank");
		}

		public event EventHandler Created;

		/// <summary>
		/// This method should be called only by LifeSpanHandler
		/// </summary>
		/// <param name="browser"></param>
		internal void OnCreated(CefBrowser browser)
		{
			// Throws exception when clicked link creates a new window
			//if (_created) throw new InvalidOperationException("Browser already created.");
			//_created = true;
			_browser = browser;

			var devToolsUrl = browser.GetHost().GetDevToolsUrl(true);
			Debug.WriteLine("DevTools URL : " + devToolsUrl);
			if (devToolsUrl != null)
			{
				Process.Start(devToolsUrl);
			}

			var handler = Created;
			if (handler != null)
			{
				handler(this, EventArgs.Empty);
			}
		}

		public event EventHandler Loaded;

		/// <summary>
		/// This method should be called only by DisplayHandler
		/// </summary>
		internal void OnLoaded()
		{
			var handler = Loaded;
			if (handler != null)
			{
				handler(this, EventArgs.Empty);
			}
		}

		internal void Close()
		{
			if (_browser != null)
			{
				_browser.Dispose();
				_browser = null;
			}
		}

		public string Url
		{
			get { return _browser.GetMainFrame().Url; }
		}
	}
}
