﻿using System;
using System.Windows.Forms;
using Xilium.CefGlue;

namespace Cef3Test
{
	public class CefPlatform : IDisposable
	{
		public int Run(string[] args)
		{
			CefRuntime.Load();

			var settings = new CefSettings
				{
					// NOTE: Disable Visual Studio hosting process for enabling multi process
					//SingleProcess = true,
					MultiThreadedMessageLoop = CefRuntime.Platform == CefRuntimePlatform.Windows,
					ReleaseDCheckEnabled = true,
					LogSeverity = CefLogSeverity.Default,
					//LogFile = "cef.log",
				};

			var argv = args;
			if (CefRuntime.Platform != CefRuntimePlatform.Windows)
			{
				argv = new string[args.Length + 1];
				Array.Copy(args, 0, argv, 1, args.Length);
				argv[0] = "-";
			}

			var mainArgs = new CefMainArgs(argv);
			var app = new CefMainApp();

			var exitCode = CefRuntime.ExecuteProcess(mainArgs, app);
			Console.WriteLine("CefRuntime.ExecuteProcess() returns {0}", exitCode);
			if (exitCode != -1)
				return exitCode;

			// guard if something wrong
			foreach (var arg in args) { if (arg.StartsWith("--type=")) { return -2; } }

			CefRuntime.Initialize(mainArgs, settings, app);

			PlatformInitialize();

			Form1 frm = new Form1(this);
			frm.Visible = true;

			PlatformRunMessageLoop();
			
			CefRuntime.Shutdown();

			PlatformShutdown();

			return 0;
		}

		private void PlatformInitialize()
		{
			Application.EnableVisualStyles();
			Application.SetCompatibleTextRenderingDefault(false);
		}

		private void PlatformShutdown()
		{
		}

		private void PlatformRunMessageLoop()
		{
			Application.Run();
		}

		private void PlatformQuitMessageLoop()
		{
			Application.Exit();
		}

		public void Quit()
		{
			PlatformQuitMessageLoop();
		}

		#region IDisposable

        ~CefPlatform()
        {
            Dispose(false);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
        }

        #endregion
	}
}
